/*********************************************************************
   Semafor EA
   Coder : Simon Anderson
   https://www.facebook.com/simon.anderson.9887/
   https://www.mql5.com/en/users/f4xtrd
   f4xtrd

   Version
   1.04  :  + Add plan B feature
   1.03  :  + Add scale in entry model
   1.02  :  + 2 re-entry type for new entry model
            + Add option to remove pending order when price already
              move X pip from the price
            + Changes in default recovery plan. On 1st losses, it will
              set next RR to 2 (and increase +1 afterward). Before it
              will increase +1RR from the latest order RR
            + New entry model. Using lower timeframe CHOC/BOS after
              price close above/lower main timeframe buy/sell level
            + Refactor line validity
   1.01  :  + Refactor CHOC detection module
              Now it can used on anytimeframe, CHOC and BOS line are
              pass as reference value in the argument
   1.00  :  + Version 1.00 final
   0.08  :  + Delete pending order if not align with CHOC direction
   0.07  :  + Delete older pending order if there's new setup
   0.06  :  + Re-entery immediately after a position is stopped out
            + Remove amount of bar for CHOC lookback, now it scan all
              bar
   0.05  :  + Semafor level 2 can be consider valid is there's
              oppposite semafore level 3
            + Once there's CHOC down and sell order setup, close the buy pending order
              since it went to the sell setup and did not trigger
            + CHOC break should be by candle close, not price touch
            + Trade intitation only check candle close, not checking candle
              open anymore
   0.04  :  + Add option to trade/not trade on CHOC level
   0.03  :  + Add default recovery plan
            + Add plan A recovery plan
   0.02  :  + Add choc detection based on last x bar
            + Add sl by fractal
   0.01  :  -
*********************************************************************/
#define VERSION "1.04"
#property copyright "Ver "+VERSION+" \nCoder - f4xtrd"
#property link "https://www.mql5.com/en/users/f4xtrd"
#property version VERSION
#property strict

#define PREFIX "SemaforEA_"
#define INDICATOR_NAME "3_Level_ZZ_Semafor"

#define BULLISH 1
#define BEARISH 2
#define BID MarketInfo(_Symbol,MODE_BID)
#define ASK MarketInfo(_Symbol,MODE_ASK)
#define BUY    0x00001
#define SELL   0x00002

struct ST_NewEntryModelVar
{
   bool              isFindingReEntry;
   bool              isCheckingLTF;
   bool              isLookingForBullishSetup;
   bool              isLookingForBearishSetup;
   double            currTFCHOCPrice;
   int               CHOC_Direction;
   double            CHOCPrice, UpperBOS, LowerBOS;
   void              ST_NewEntryModelVar()
   {
      isFindingReEntry = false;
      isCheckingLTF = false;
      isLookingForBullishSetup = false;
      isLookingForBearishSetup = false;
      currTFCHOCPrice = -1;
      CHOC_Direction = -1;
      CHOCPrice = 0;
      UpperBOS = 0;
      LowerBOS = 0;
   };
};

struct ST_CheckNewBarRecord
{
   int               timeframe;
   datetime          lastCheck;
   void              ST_CheckNewBarRecord(int p_timeframe) : timeframe(p_timeframe)
   {
      lastCheck = iTime(_Symbol, timeframe, 0);
   }
   /*----*/
   bool              isNewBar()
   {
      datetime timeNow = iTime(_Symbol, timeframe, 1);
      if (lastCheck == timeNow) return false;
      lastCheck = iTime(_Symbol, timeframe, 1);
      return true;
   }
};

struct ST_OrderCount
{
   int               buy;
   int               sell;
   int               buystop;
   int               sellstop;
   int               buylimit;
   int               selllimit;
   void              ST_OrderCount()
   {
      buy = 0;
      sell = 0;
      buystop = 0;
      sellstop = 0;
      buylimit = 0;
      selllimit = 0;
   }
   /*----*/
   void              count(const int p_magicNumber)
   {
      buy = 0;
      sell = 0;
      buystop = 0;
      sellstop = 0;
      buylimit = 0;
      selllimit = 0;
      for(int i = 0; i < OrdersTotal(); i++)
      {
         if (!OrderSelect(i, SELECT_BY_POS)) continue;
         if (OrderSymbol() != _Symbol) continue;
         if (OrderMagicNumber() != p_magicNumber && p_magicNumber != 0) continue;
         if (OrderType() == OP_BUY)
            buy++;
         if (OrderType() == OP_SELL)
            sell++;
         if (OrderType() == OP_BUYSTOP)
            buystop++;
         if (OrderType() == OP_SELLSTOP)
            sellstop++;
         if (OrderType() == OP_BUYLIMIT)
            buylimit++;
         if (OrderType() == OP_SELLLIMIT)
            selllimit++;
      }
   }
};

struct ST_Order
{
   int               ticket;
   datetime          opentime;
   int               type;
   double            size;
   string            symbol;
   double            openprice;
   double            stoploss;
   double            takeprofit;
   datetime          closetime;
   double            closeprice;
   double            commision;
   double            swap;
   double            profit;
   int               magic;
   string            comment;
   datetime          expire;
   /*----*/
   void              ST_Order()
      :
                     ticket(NULL),
                     opentime(NULL),
                     type(NULL),
                     size(NULL),
                     symbol(""),
                     openprice(NULL),
                     stoploss(NULL),
                     takeprofit(NULL),
                     closetime(NULL),
                     closeprice(NULL),
                     commision(NULL),
                     swap(NULL),
                     profit(NULL),
                     magic(NULL),
                     comment(""),
                     expire(NULL)
   {}
};

struct ST_SendOrderResult
{
   bool              isSuccess;
   int               errCode;
   datetime          lastOPtime;
   int               ticket[];
   void              ST_SendOrderResult() : isSuccess(false), errCode(0), lastOPtime(0)
   {
      ArrayFree(ticket);
   }
};

enum eNSemaforType
{
   en_upperSemafor,
   en_lowerSemafor,
   en_none
};

struct sTSemaforDot
{
   eNSemaforType     type;
   int               period;
   double            price;
   datetime          time;
   void              sTSemaforDot(): type(en_none), period(0), price(0), time(0)
   {
   }
};

enum EN_NEW_ENTRY_MODEL_STOPLOSS
{
   en_newEntryModelStoplossByPoint, // By point
   en_newEntryModelStoplossBySemafor // By semafor
};

enum EN_ENTRY_MODEL
{
   en_entryModelOriginal, // Original entry model
   en_entryModelNew, // New entry model
   en_entryModelScaleIn // Scale in entry model
};

enum EN_NEW_ENTRY_MODEL_REENTRY_OPTION
{
   en_reentrySamePrice, // Re-entry on same price
   en_reentryNewSetup // Re-entry on new setup
};

input string prm_1 = "==================================";                                   //| ø  STRATEGY CONFIGURATION
input EN_ENTRY_MODEL i_EntryModel = en_entryModelOriginal;                                   //|      - Entry model
input double i_RiskPerTrade = 5;                                                             //|      - Risk per trade ($)
input double i_Commision = 7;                                                                //|      - Commision (in point)
input int i_MagicNumber = 1234567890;                                                        //|      - Magic number (max 10 digit)
input bool i_TradeCHOCLevel = true;                                                          //|      - Trade CHOC level
input bool i_DeletePendingOrderIfAgainstCHOC = true;                                         //|      - Delete pending order against CHOC
input double i_RemovePO_AndLevelWhenPriceMoveFurther = 200;                                  //|      - Delete level & PO when price move (0 to disable)
input bool i_UseRecover = true;                                                              //|      - Use recovery plan
input bool i_UsePlanA = true;                                                                //|      - Use plan A
input double i_PlanA_MaxLossTrade = 5;                                                       //|         - Max loss (number of trades)
input double i_PlanANewRR = 2;                                                               //|         - Max RR
input bool i_UsePlanB = true;                                                                //|      - Use plan B
input double i_PlanBLotSize = 0.05;                                                          //|         - Lot size trigger
input string exp_3 = "[ Semafor setting ]";                                                  //|      [ Semafor Setting]
input double i_SemaforPeriod_1 = 5.0;                                                        //|         - Period 1
input double i_SemaforPeriod_2 = 13.0;                                                       //|         - Period 2
input double i_SemaforPeriod_3 = 34.0;                                                       //|         - Period 3
input string i_SemaforDevStep_1 = "1,3";                                                     //|         - Dev step 1
input string i_SemaforDevStep_2 = "8,5";                                                     //|         - Dev step 2
input string i_SemaforDevStep_3 = "21,12";                                                   //|         - Dev step 3
input bool i_UsePeriod_2 = true;                                                             //|         - Use Period 2
input bool i_UsePeriod_3 = true;                                                             //|         - Use Period 3
input int i_LookBackBar = 200;                                                               //|      - Semafor lookback bar amount
input string exp_258 = "[ Original Entry Model ]";                                           //| ø  ORIGINAL STRATERY
input string exp_262 = "[ Choose only 1. Set parameter to 0 or false to disable ]";          //|      [ Stoploss Option ]
input double i_Stoploss = 0;                                                                 //|         - Stoploss (in point)
input bool i_StoplossByZigzag = true;                                                        //|         - Stoploss by zigzag
input int i_ZigagDepth = 12;                                                                 //|            - Zigzag Depth
input int i_ZigzagDeviation = 5;                                                             //|            - Zigzag Deviation
input int i_ZigzagBacktstep = 3;                                                             //|            - Zigzag Backstep
input string exp_4 = "[ Choose only 1. Set parameter to 0 to disable ]";                     //|      [ Takeprofit option ]
input double i_TakeProfitRR = 2;                                                             //|         - Takeprofit (RR)
input string exp_215 = "[ New Entry Model ]";                                                //| ø  NEW ENTRY STRATEGY
input ENUM_TIMEFRAMES i_NewEntryModelTimeframe = PERIOD_M5;                                  //|      - Lower timeframe
input EN_NEW_ENTRY_MODEL_REENTRY_OPTION i_NewEntryModelReentryOption = en_reentrySamePrice;  //|      - Re-entry type
input string exp_225 = "[ Choose only 1. Set parameter to 0 or false to disable ]";             //|      [ Stoploss Option ]
input EN_NEW_ENTRY_MODEL_STOPLOSS i_NewEntryModelStoploss = en_newEntryModelStoplossBySemafor;  //|         - Stoploss type
input double i_NewEntryStoplossPoint = 200;                                                     //|         - Stoploss (in point)
input string exp_273 = "[ Scale In Entry Model ]";                                           //| ø  SCALE IN STRATEGY
input double i_LotSize = 0.01;                                                               //|      - Lot size
input double i_TakeProfitBasket = 500;                                                       //|      - Take profit basket
input double i_MaxDistanceToCHOC = 800;                                                      //|      - Max distance to CHOC (0 to disable)
input int i_CutLossAfterCHOC = -500;                                                         //|      - Cut loss point after CHOC

/*----*/
string i_Comment = "";                                                     //|      - Comment (max 32 char)
int i_MaxSlippage = 50;                                                    //|      - Max slippage
int i_MaxRetries = 3;                                                      //|      - Max open order retries
int g_chochDirection = -1;
/*----*/
sTSemaforDot g_checkedSemafor[];
int g_managedPosition[];
double g_chocLevel = 0;
datetime g_chocTime = 0;
double g_startingEquity = AccountEquity();
int g_lastPosition = 0;
string entryModel[] = { "Original Entry Model", "New Entry Model", "Scale in Entry Model" };
string reEntryModel[] = { "Re-entry on same price", "Re-entry on new setup" };

//+------------------------------------------------------------------+
int OnInit()
{
   ArrayFree(g_checkedSemafor);
   scanSemafor(i_LookBackBar);
   checkLineValidity();
   if (isInputInvalid())
      return INIT_FAILED;
   return INIT_SUCCEEDED;
}

//+------------------------------------------------------------------+
void OnDeinit(const int reason)
{
   ObjectsDeleteAll(ChartID(), PREFIX);
}

//+------------------------------------------------------------------+
void OnTick()
{
   if (i_EntryModel == en_entryModelNew)
      newEntryModel();
   if (i_EntryModel == en_entryModelOriginal)
      originalEntryModel();
   if (i_EntryModel == en_entryModelScaleIn)
      scaleInEntryModel();
}

//+------------------------------------------------------------------+
void scaleInEntryModel()
{
   static ST_CheckNewBarRecord chart(PERIOD_CURRENT);
   checkPriceMoveAwayFromPO();
   checkTotalFloatingLoss();
   if (!chart.isNewBar())
      return;
   checkPendingOrderDirectionToCHOCH();
   checkNewSemafor();
   checkLineRemovalCondition();
   checkTradingCondition_ScaleInEntryModel();
   extendLine();
   checkPositionBasketTP();
   /*-- check choc direction after all other module --*/
   double upper_bos = 0, lower_bos = 0;
   getCHOC_Direction(PERIOD_CURRENT, g_chochDirection, g_chocLevel, g_chocTime, upper_bos, lower_bos);
   drawCHOC();
   showInfo();
}

//+------------------------------------------------------------------+
void drawCHOC()
{
   createTrendLine(ChartID(), "choc_line", g_chocLevel, g_chocLevel, g_chocTime, iTime(_Symbol, _Period, 0) + PeriodSeconds(PERIOD_CURRENT) * 20, clrGray, 0, 5, true, true);
}

//+------------------------------------------------------------------+
void newEntryModel()
{
   static ST_NewEntryModelVar newEntryModelVar;
   static ST_CheckNewBarRecord lftChart(i_NewEntryModelTimeframe);
   static ST_CheckNewBarRecord currChart(PERIOD_CURRENT);
   /*----*/
   checkPositionHitTP();
   /*----*/
   checkReEntry_NewEntryModel(newEntryModelVar);
   if (lftChart.isNewBar())
      if (newEntryModelVar.isCheckingLTF || newEntryModelVar.isFindingReEntry)
         updateLTF_Level(newEntryModelVar);
   if (newEntryModelVar.isLookingForBearishSetup || newEntryModelVar.isLookingForBullishSetup)
      checkPriceTrigger(newEntryModelVar);
   if (!currChart.isNewBar())
      return;
   /*----*/
   checkNewSemafor();
   checkLineRemovalCondition();
   extendLine();
   /*-- check if there's candle close above below buy/sell level --*/
   checkTradingCondition_NewEntryModel(newEntryModelVar);
   /*-- get current TF choc direction --*/
   double upper_bos = 0, lower_bos = 0;
   getCHOC_Direction(PERIOD_CURRENT, newEntryModelVar.CHOC_Direction, newEntryModelVar.currTFCHOCPrice, g_chocTime, upper_bos, lower_bos);
   drawCHOC();
   showNewEntryInfo(newEntryModelVar);
}

//+------------------------------------------------------------------+
void checkPriceTrigger(ST_NewEntryModelVar &p_var)
{
   static double lastBid = 0;
   double currentBid = MarketInfo(_Symbol, MODE_BID);
   /*-- bullish direction --*/
   if (p_var.CHOC_Direction == BULLISH && p_var.isCheckingLTF && p_var.isLookingForBullishSetup)
   {
      if (  (currentBid >= p_var.CHOCPrice && lastBid < p_var.CHOCPrice)
            ||
            (currentBid >= p_var.UpperBOS && lastBid < p_var.UpperBOS)
         )
      {
         ST_SendOrderResult result = openTrade_NewEntryModel(BULLISH);
         if (result.isSuccess && ArraySize(result.ticket) > 0)
         {
            /*-- add to managed position array --*/
            addToEndOfArray(g_managedPosition, result.ticket[0]);
            g_lastPosition = result.ticket[0];
            /*-- turn off flag to search at LTF ---*/
            p_var.isCheckingLTF = false;
            p_var.isLookingForBullishSetup = false;
         }
      }
   }
   /*-- bearish direction --*/
   if (p_var.CHOC_Direction == BEARISH && p_var.isCheckingLTF && p_var.isLookingForBearishSetup)
   {
      if (  (currentBid <= p_var.CHOCPrice && lastBid > p_var.CHOCPrice)
            ||
            (currentBid <= p_var.LowerBOS && lastBid > p_var.LowerBOS)
         )
      {
         ST_SendOrderResult result = openTrade_NewEntryModel(BEARISH);
         if (result.isSuccess && ArraySize(result.ticket) > 0)
         {
            /*-- add to managed position array --*/
            addToEndOfArray(g_managedPosition, result.ticket[0]);
            g_lastPosition = result.ticket[0];
            /*-- turn off flag to search at LTF ---*/
            p_var.isCheckingLTF = false;
            p_var.isLookingForBearishSetup = false;
         }
      }
   }
   lastBid = currentBid;
}

//+------------------------------------------------------------------+
ST_SendOrderResult openTrade_NewEntryModel(int p_direction)
{
   ST_SendOrderResult sendOrderResult;
   ST_Order order;
   order.comment = "";
   if (i_UsePlanB && isPlanB_Active())
      order.comment += order.comment + "#B";
   order.magic = i_MagicNumber;
   order.openprice = p_direction == BULLISH ? ASK : BID;
   order.type = p_direction == BULLISH ? OP_BUY : OP_SELL;
   order.stoploss = getNewEntryModelStoploss(order.type, order.openprice);
   double slLength = MathAbs(order.openprice - order.stoploss);
   double tpRR = getTakeProfitRR();
   if (i_TakeProfitRR != 0)
      order.takeprofit = (order.type == OP_BUYSTOP || order.type == OP_BUYLIMIT || order.type == OP_BUY) ? order.openprice + (tpRR * slLength)
                         : order.openprice - (tpRR * slLength);
   order.size = getLotSize(slLength);
   order.openprice = NormalizeDouble(order.openprice, _Digits);
   order.stoploss = NormalizeDouble(order.stoploss, _Digits);
   order.takeprofit = NormalizeDouble(order.takeprofit, _Digits);
   sendOrderResult = sendOrder(order, i_MaxSlippage, clrLimeGreen, i_MaxRetries);
   return sendOrderResult;
}

//+------------------------------------------------------------------+
double getNewEntryModelStoploss(int p_type, double p_orderOpenPrice)
{
   double result = 0;
   if (p_type == OP_BUYSTOP || p_type == OP_BUYLIMIT || p_type == OP_BUY)
   {
      if (i_NewEntryModelStoploss == en_newEntryModelStoplossBySemafor)
         result = getLatestSemafore(MODE_LOWER);
      if (i_NewEntryModelStoploss == en_newEntryModelStoplossByPoint)
         result = p_orderOpenPrice - (i_NewEntryStoplossPoint * _Point);
   }
   /*----*/
   if (p_type == OP_SELLSTOP || p_type == OP_SELLLIMIT || p_type == OP_SELL)
   {
      if (i_NewEntryModelStoploss == en_newEntryModelStoplossBySemafor)
         result = getLatestSemafore(MODE_UPPER);
      if (i_NewEntryModelStoploss == en_newEntryModelStoplossByPoint)
         result = p_orderOpenPrice + (i_NewEntryStoplossPoint * _Point);
   }
   return result;
}

//+------------------------------------------------------------------+
double getLatestSemafore(int p_type)
{
   for(int i = 1; i < iBars(_Symbol, i_NewEntryModelTimeframe); i++)
   {
      sTSemaforDot currSemafor = getSemaforDot(i_NewEntryModelTimeframe, i);
      if (currSemafor.price != 0)
      {
         if (  (p_type == MODE_LOWER && currSemafor.type == en_lowerSemafor)
               ||
               (p_type == MODE_UPPER && currSemafor.type == en_upperSemafor)
            )
            return currSemafor.price;
      }
   }
   return 0;
}

//+------------------------------------------------------------------+
void checkReEntry_NewEntryModel(ST_NewEntryModelVar &p_var)
{
   if (i_NewEntryModelReentryOption == en_reentrySamePrice)
   {
      static bool isReEntrySameLevelActive = false;
      static double s_reEntryPrice = 0;
      static int s_reEntryType = 0;
      static double s_lastBid = 0;
      double currentBid = MarketInfo(_Symbol, MODE_BID);
      /*-- if g_lastPosition hit sl & reenterSamelevelactive = false -> get the entry price and op type --*/
      if (  g_lastPosition != 0  )
      {
         if (  !OrderSelect(g_lastPosition, SELECT_BY_TICKET)  )
            return;
         if (  OrderCloseTime() > 0
               && isStringFind(OrderComment(), "[sl]")
            )
         {
            s_reEntryPrice = OrderOpenPrice();
            s_reEntryType = OrderType();
            g_lastPosition = 0;
            /*-- active re-entry same level flag --*/
            isReEntrySameLevelActive = true;
            /*-- flag so other module don't open order on new setup --*/
            p_var.isFindingReEntry = true;
         }
      }
      /*----*/
      if (p_var.isFindingReEntry && isReEntrySameLevelActive && s_reEntryPrice != 0)
      {
         /*--check condition that stop re-entry -> CHOC --*/
         if (  (s_reEntryType == OP_BUY && p_var.CHOC_Direction == BEARISH )
               ||
               (s_reEntryType == OP_SELL && p_var.CHOC_Direction == BULLISH)  )
         {
            p_var.isFindingReEntry = false;
            return;
         }
         /*-- check price pass the entry level -> open trade -*/
         if (  s_reEntryType == OP_BUY
               && currentBid >= s_reEntryPrice
               && s_lastBid < s_reEntryPrice
               && p_var.CHOC_Direction == BULLISH
            )
         {
            ST_SendOrderResult result = openTrade_NewEntryModel(BULLISH);
            if (result.isSuccess && ArraySize(result.ticket) > 0)
            {
               /*-- add to managed position array --*/
               addToEndOfArray(g_managedPosition, result.ticket[0]);
               g_lastPosition = result.ticket[0];
               isReEntrySameLevelActive = false;
               p_var.isFindingReEntry = false;
            }
         }
         /*-- check price pass the entry level -> open trade -*/
         if (  s_reEntryType == OP_SELL
               && currentBid <= s_reEntryPrice
               && s_lastBid > s_reEntryPrice
               && p_var.CHOC_Direction == BEARISH
            )
         {
            ST_SendOrderResult result = openTrade_NewEntryModel(BEARISH);
            if (result.isSuccess && ArraySize(result.ticket) > 0)
            {
               /*-- add to managed position array --*/
               addToEndOfArray(g_managedPosition, result.ticket[0]);
               g_lastPosition = result.ticket[0];
               isReEntrySameLevelActive = false;
               p_var.isFindingReEntry = false;
            }
         }
      }
      s_lastBid = currentBid;
   }
   /*----*/
   if (i_NewEntryModelReentryOption == en_reentryNewSetup)
   {
      static int s_reEntryType = 0;
      static double s_lastBid = 0;
      double currentBid = MarketInfo(_Symbol, MODE_BID);
      /*-- check g_last position --*/
      if (  g_lastPosition != 0
            && OrderSelect(g_lastPosition, SELECT_BY_TICKET))
      {
         if (  OrderCloseTime() > 0
               && isStringFind(OrderComment(), "[sl]"))
         {
            /*-- set flag to check for LTF to true --*/
            p_var.isFindingReEntry = true;
            s_reEntryType = OrderType();
            g_lastPosition = 0;
         }
      }
      /*----*/
      if (p_var.isFindingReEntry)
      {
         /*--check condition that stop re-entry -> CHOC --*/
         if (  (s_reEntryType == OP_BUY && p_var.CHOC_Direction == BEARISH )
               ||
               (s_reEntryType == OP_SELL && p_var.CHOC_Direction == BULLISH)  )
         {
            p_var.isFindingReEntry = false;
            return;
         }
         /*----*/
         bool isBuyLevelTriggered = (currentBid >= p_var.UpperBOS && s_lastBid < p_var.UpperBOS)
                                    ||
                                    (currentBid >= p_var.CHOCPrice && s_lastBid < p_var.CHOCPrice);
         bool isSellLevelTriggered = (currentBid <= p_var.LowerBOS && s_lastBid > p_var.LowerBOS)
                                     ||
                                     (currentBid <= p_var.CHOCPrice && s_lastBid > p_var.CHOCPrice);
         /*-- check price pass the entry level -> open trade -*/
         if (  s_reEntryType == OP_BUY
               && isBuyLevelTriggered
               && p_var.CHOC_Direction == BULLISH
            )
         {
            ST_SendOrderResult result = openTrade_NewEntryModel(BULLISH);
            if (result.isSuccess && ArraySize(result.ticket) > 0)
            {
               /*-- add to managed position array --*/
               addToEndOfArray(g_managedPosition, result.ticket[0]);
               g_lastPosition = result.ticket[0];
               p_var.isFindingReEntry = false;
            }
         }
         /*----*/
         if (  s_reEntryType == OP_SELL
               && isSellLevelTriggered
               && p_var.CHOC_Direction == BEARISH
            )
         {
            ST_SendOrderResult result = openTrade_NewEntryModel(BEARISH);
            if (result.isSuccess && ArraySize(result.ticket) > 0)
            {
               /*-- add to managed position array --*/
               addToEndOfArray(g_managedPosition, result.ticket[0]);
               g_lastPosition = result.ticket[0];
               p_var.isFindingReEntry = false;
            }
         }
      }
      /*----*/
      s_lastBid = currentBid;
   }
}

//+------------------------------------------------------------------+
void checkTradingCondition_NewEntryModel(ST_NewEntryModelVar &p_var)
{
   ST_OrderCount orderTotal;
   orderTotal.count(i_MagicNumber);
   for(int i = ObjectsTotal(ChartID()) - 1; i >= 0; i--)
   {
      string name = ObjectName(ChartID(), i);
      if (ObjectGetInteger(ChartID(), name, OBJPROP_TYPE) != OBJ_TREND)
         continue;
      double price = ObjectGetDouble(ChartID(), name, OBJPROP_PRICE1);
      bool isAllowToTradeOnLevel = i_TradeCHOCLevel ||
                                   (!i_TradeCHOCLevel && NormalizeDouble(price, _Digits) != NormalizeDouble(p_var.currTFCHOCPrice, _Digits));
      /*-- if EA trying to re-entry and mode is re-entry on same price -> not allowed to open on new setup --*/
      bool isNotReEntryingPhase = !p_var.isFindingReEntry;
      if (  isStringFind(name, "upper")
            && iClose(_Symbol, _Period, 1) > price
            && (orderTotal.buy + orderTotal.sell <= 0)
            && p_var.CHOC_Direction == BEARISH
            && isAllowToTradeOnLevel
            && isNotReEntryingPhase
         )
      {
         p_var.isCheckingLTF = true;
         p_var.isLookingForBearishSetup = true;
         p_var.isLookingForBullishSetup = false;
      }
      /*----*/
      if (  isStringFind(name, "lower")
            && iClose(_Symbol, _Period, 1) < price
            && (orderTotal.buy + orderTotal.sell <= 0)
            && p_var.CHOC_Direction == BULLISH
            && isAllowToTradeOnLevel
            && isNotReEntryingPhase
         )
      {
         p_var.isCheckingLTF = true;
         p_var.isLookingForBullishSetup = true;
         p_var.isLookingForBearishSetup = false;
      }
   }
   /*-- to deactive looking for setup when there's directional CHOC --*/
   if (p_var.isLookingForBearishSetup && p_var.CHOC_Direction == BULLISH)
   {
      p_var.isLookingForBearishSetup = false;
      p_var.isCheckingLTF = false;
   }
   if (p_var.isLookingForBullishSetup && p_var.CHOC_Direction == BEARISH)
   {
      p_var.isLookingForBullishSetup = false;
      p_var.isCheckingLTF = false;
   }
}

//+------------------------------------------------------------------+
void updateLTF_Level(ST_NewEntryModelVar &p_var)
{
   /*-- find upper lower semafor on lower TF --*/
   sTSemaforDot upperSemafore, lowerSemafore;
   for(int i = 1; i < iBars(_Symbol, i_NewEntryModelTimeframe); i++)
   {
      sTSemaforDot currSemafor = getSemaforDot(i_NewEntryModelTimeframe, i);
      if (  currSemafor.type == en_lowerSemafor
            && lowerSemafore.price == 0
            && (currSemafor.period == 3 || currSemafor.period == 2)
            && isFoundOppositeSemafor(i_NewEntryModelTimeframe, i, en_upperSemafor, currSemafor.period)
         )
         lowerSemafore = currSemafor;
      if (  currSemafor.type == en_upperSemafor
            && upperSemafore.price == 0
            && (currSemafor.period == 3 || currSemafor.period == 2)
            && isFoundOppositeSemafor(i_NewEntryModelTimeframe, i, en_lowerSemafor, currSemafor.period)
         )
         upperSemafore = currSemafor;
      if (upperSemafore.price != 0 && lowerSemafore.price != 0)
         break;
   }
   p_var.CHOCPrice = 0;
   p_var.UpperBOS = upperSemafore.price;
   p_var.LowerBOS = lowerSemafore.price;
}

//+------------------------------------------------------------------+
void originalEntryModel()
{
   static ST_CheckNewBarRecord chart(PERIOD_CURRENT);
   checkPriceMoveAwayFromPO();
   checkReEntry();
   if (!chart.isNewBar())
      return;
   checkPendingOrderDirectionToCHOCH();
   checkNewSemafor();
   checkLineRemovalCondition();
   checkTradingCondition_OriginalEntryModel();
   extendLine();
   checkPositionHitTP();
   /*-- check choc direction after all other module --*/
   double upper_bos = 0, lower_bos = 0;
   getCHOC_Direction(PERIOD_CURRENT, g_chochDirection, g_chocLevel, g_chocTime, upper_bos, lower_bos);
   drawCHOC();
   showInfo();
}

//+------------------------------------------------------------------+
void checkPriceMoveAwayFromPO()
{
   if (i_RemovePO_AndLevelWhenPriceMoveFurther == 0)
      return;
   /*----*/
   for(int i = OrdersTotal() - 1; i >= 0; i--)
   {
      if (OrderSelect(i, SELECT_BY_POS) == false)
         continue;
      if (OrderSymbol() != _Symbol || OrderMagicNumber() != i_MagicNumber)
         continue;
      if (  (OrderType() == OP_BUYSTOP || OrderType() == OP_BUYLIMIT)
            && MathAbs(OrderOpenPrice() - BID) > i_RemovePO_AndLevelWhenPriceMoveFurther * _Point)
      {
         bool result = OrderDelete(OrderTicket(), clrNONE);
         double price = StringToDouble(OrderComment());
         deleteLine(price);
      }
      /*----*/
      if (  (OrderType() == OP_SELLSTOP || OrderType() == OP_SELLLIMIT)
            && MathAbs(OrderOpenPrice() - BID) > i_RemovePO_AndLevelWhenPriceMoveFurther * _Point)
      {
         bool result = OrderDelete(OrderTicket(), clrNONE);
         double price = StringToDouble(OrderComment());
         deleteLine(price);
      }
   }
}

//+------------------------------------------------------------------+
void deleteLine(double p_price)
{
   for(int i = ObjectsTotal(ChartID()); i >= 0; i--)
   {
      string name = ObjectName(ChartID(), i);
      double price_1 = ObjectGetDouble(0, name, OBJPROP_PRICE1);
      if (NormalizeDouble(price_1, _Digits) == NormalizeDouble(p_price, _Digits))
         ObjectDelete(0, name);
   }
}

//+------------------------------------------------------------------+
// Calculate lot based on risk, stoploss length and symbol
double calculateOrderLot
(
   const double p_balance,
   const double p_riskPercent, // In percentage
   const double p_stoplossLength, // In point/pip size
   const string p_symbol,
   bool p_roundToMin = true
)
{
   double result;
   int OrdersizeAllowed = 0;
   if(MarketInfo(p_symbol, MODE_MINLOT) == 0.01) OrdersizeAllowed = 2;
   if(MarketInfo(p_symbol, MODE_MINLOT) == 0.1) OrdersizeAllowed = 1;
   double p_riskDecimal = p_riskPercent / 100;
   result = p_balance * p_riskDecimal / p_stoplossLength / (MarketInfo(p_symbol, MODE_TICKVALUE));
   if (result < MarketInfo(p_symbol, MODE_MINLOT) && p_roundToMin) result = MarketInfo(p_symbol, MODE_MINLOT);
   return NormalizeDouble(result, OrdersizeAllowed);
};

//+------------------------------------------------------------------+
void checkPendingOrderDirectionToCHOCH()
{
   if (!i_DeletePendingOrderIfAgainstCHOC)
      return;
   for(int i = OrdersTotal() - 1; i >= 0; i--)
   {
      if (OrderSelect(i, SELECT_BY_POS) == false)
         continue;
      if (OrderSymbol() != _Symbol || OrderMagicNumber() != i_MagicNumber)
         continue;
      if ((OrderType() == OP_BUYSTOP || OrderType() == OP_BUYLIMIT) && g_chochDirection == BEARISH)
         closeOrderByTicket(OrderTicket(), i_MaxSlippage, clrNONE, i_MaxRetries);
      if ((OrderType() == OP_SELLSTOP || OrderType() == OP_SELLLIMIT) && g_chochDirection == BULLISH)
         closeOrderByTicket(OrderTicket(), i_MaxSlippage, clrNONE, i_MaxRetries);
   }
}

//+------------------------------------------------------------------+
void checkReEntry()
{
   if (g_lastPosition == 0)
      return;
   if (OrderSelect(g_lastPosition, SELECT_BY_TICKET) == false)
      return;
   if (OrderCloseTime() > 0 && isStringFind(OrderComment(), "[sl]"))
   {
      double openPrice = OrderOpenPrice();
      int type = OrderType() == OP_BUY ? BUY : SELL;
      double stoploss = OrderStopLoss();
      if (  (OrderType() == OP_BUY && g_chochDirection == BEARISH)
            ||
            (OrderType() == OP_SELL && g_chochDirection == BULLISH)
         )
      {
         g_lastPosition = 0;
         return;
      }
      ST_SendOrderResult sendResult = openReEntryPendingOrder(type, openPrice, stoploss);
      if (sendResult.isSuccess && ArraySize(sendResult.ticket) > 0)
      {
         /*-- if (isPlanB_Active())
            g_planB_Ctr++; --*/
         g_lastPosition = sendResult.ticket[0];
         addToEndOfArray(g_managedPosition, sendResult.ticket[0]);
      }
   }
}

//+------------------------------------------------------------------+
void showInfo()
{
#define RET "\n"
#define SPC "   "
   string msg = "";
   string curr = AccountInfoString(ACCOUNT_CURRENCY);
   ST_OrderCount total = getOrderCount(i_MagicNumber);
   msg = msg + RET + RET;
   msg = msg + SPC + "Semafor EA Ver." + VERSION + RET;
   msg = msg + SPC + "---------------------------------------" + RET;
   msg = msg + SPC + "Entry model = " + entryModel[i_EntryModel] + RET;
   msg = msg + SPC + "CHOC direction = " + (g_chochDirection == BULLISH ? "BULLISH" : "BEARISH") + RET;
   msg = msg + SPC + "---------------------------------------" + RET;
   msg = msg + SPC + "In current cyle : " + RET;
   msg = msg + SPC + "Total position loss = " + (string)getTotalPositionLossCount() + RET;
   msg = msg + SPC + "Total dollar loss = " + DoubleToStr(getTotalPositionDollarLoss(), 2) + RET;
   msg = msg + SPC + "Total closed lot = " + (string)getTotalClosedLot() + RET;
   msg = msg + SPC + "Total position opened in B plan = " + DoubleToStr(getTotalPlanBOpenedOrder(), 0) + RET;
   Comment(msg);
}

//+------------------------------------------------------------------+
void showNewEntryInfo(ST_NewEntryModelVar &p_var)
{
   string msg = "";
   string curr = AccountInfoString(ACCOUNT_CURRENCY);
   ST_OrderCount total = getOrderCount(i_MagicNumber);
   msg = msg + RET + RET;
   msg = msg + SPC + "Semafor EA Ver." + VERSION + RET;
   msg = msg + SPC + "---------------------------------------" + RET;
   msg = msg + SPC + "Entry model = " + entryModel[i_EntryModel] + RET;
   msg = msg + SPC + "Re-entry model = " + reEntryModel[i_NewEntryModelReentryOption] + RET;
   msg = msg + SPC + "CHOC direction = " + (p_var.CHOC_Direction == BULLISH ? "BULLISH" : "BEARISH") + RET;
   msg = msg + SPC + "Looking LTF = " + (p_var.isCheckingLTF ? "true" : "false") + RET;
   msg = msg + SPC + "---------------------------------------" + RET;
   msg = msg + SPC + "In current cyle : " + RET;
   msg = msg + SPC + "Total position loss = " + (string)getTotalPositionLossCount() + RET;
   msg = msg + SPC + "Total dollar loss = " + DoubleToStr(getTotalPositionDollarLoss(), 2) + RET;
   msg = msg + SPC + "Total closed lot = " + (string)getTotalClosedLot() + RET;
   msg = msg + SPC + "Total position opened in B plan = " + DoubleToStr(getTotalPlanBOpenedOrder(), 0) + RET;
   Comment(msg);
}

//+------------------------------------------------------------------+
void checkPositionHitTP()
{
   if (isPositionHitTP())
   {
      ArrayFree(g_managedPosition);
      g_startingEquity = AccountEquity();
      closeAllOrder(i_MagicNumber, i_MaxSlippage);
   }
}

//+------------------------------------------------------------------+
void checkPositionBasketTP()
{
   double totalPL = 0;
   for(int i = OrdersTotal() - 1; i >= 0; i--)
   {
      if (OrderSelect(i, SELECT_BY_POS) == false)
         continue;
      if (OrderSymbol() != _Symbol || OrderMagicNumber() != i_MagicNumber)
         continue;
      if (OrderType() > 1)
         continue;
      totalPL += OrderProfit();
      totalPL += OrderCommission();
      totalPL += OrderSwap();
   }
   /*----*/
   double tickValue = MarketInfo(OrderSymbol(), MODE_TICKVALUE),
          point = MarketInfo(OrderSymbol(), MODE_POINT);
   double tpInDollar = i_LotSize * tickValue * i_TakeProfitBasket;
   double closedPL = getTotalBasketDollarClosedPL();
   if (  totalPL >= tpInDollar + MathAbs(closedPL)  )
   {
      /*--#del--*/
      PrintFormat("### basket tp reached. Close all order");
      closeAllOrder(i_MagicNumber, i_MaxSlippage);
      ArrayFree(g_managedPosition);
      g_lastPosition = 0;
   }
}

//+------------------------------------------------------------------+
double getTotalBasketDollarClosedPL()
{
   double total = 0;
   for(int i = ArraySize(g_managedPosition) - 1; i >= 0; i--)
   {
      if (OrderSelect(g_managedPosition[i], SELECT_BY_TICKET) == false)
         continue;
      if (OrderCloseTime() <= 0)
         continue;
      total += OrderProfit();
      total += OrderCommission();
      total += OrderSwap();
   }
   return total;
}

//+------------------------------------------------------------------+
void checkTotalFloatingLoss()
{
   /*-- check total floating loss & trend change -> close all --*/
   double totalFloatingPL_Point = 0;
   int tradeDirection = -1;
   for(int i = OrdersTotal() - 1; i >= 0; i--)
   {
      if (OrderSelect(i, SELECT_BY_POS) == false)
         continue;
      if (OrderSymbol() != _Symbol || OrderMagicNumber() != i_MagicNumber || OrderType() > 1)
         continue;
      double plPoint = OrderType() == OP_BUY ? BID - OrderOpenPrice() : OrderOpenPrice() - ASK;
      totalFloatingPL_Point += plPoint;
      if (tradeDirection == -1)
         tradeDirection = OrderType();
   }
   /*----*/
   bool isTradeDirectionAgainstCHOC = (tradeDirection == OP_BUY && g_chochDirection == BEARISH )
                                      ||
                                      (tradeDirection == OP_SELL && g_chochDirection == BULLISH);
   if (totalFloatingPL_Point < i_CutLossAfterCHOC * _Point && isTradeDirectionAgainstCHOC)
   {
      closeAllOrder(i_MagicNumber, i_MaxSlippage);
   }
}


//+------------------------------------------------------------------+
bool isPositionHitTP()
{
   bool result = false;
   for(int i = ArraySize(g_managedPosition) - 1; i >= 0; i--)
   {
      if (OrderSelect(g_managedPosition[i], SELECT_BY_TICKET) == false)
         continue;
      if (OrderCloseTime() > 0)
      {
         if (!isStringFind(OrderComment(), "[tp]"))
            continue;
         result = true;
         break;
      }
   }
   return result;
}

//+------------------------------------------------------------------+
int getTotalPositionLossCount()
{
   int counter = 0;
   for(int i = ArraySize(g_managedPosition) - 1; i >= 0; i--)
   {
      if (OrderSelect(g_managedPosition[i], SELECT_BY_TICKET) == false)
         continue;
      if (OrderCloseTime() > 0 && isStringFind(OrderComment(), "[sl]"))
         counter++;
   }
   return counter;
}

//+------------------------------------------------------------------+
bool isInputInvalid()
{
   if (i_StoplossByZigzag && i_Stoploss != 0)
   {
      Alert("Invalid stoploss input. Please choose only 1.");
      return true;
   }
   /*----*/
   return false;
}

//+------------------------------------------------------------------+
void extendLine()
{
   for(int i = ObjectsTotal(ChartID()); i >= 0; i--)
   {
      string objName = ObjectName(ChartID(), i);
      if (!isStringFind(objName, PREFIX))
         continue;
      if (ObjectGetInteger(ChartID(), objName, OBJPROP_TYPE) != OBJ_TREND)
         continue;
      ObjectSetInteger(ChartID(), objName, OBJPROP_TIME2, iTime(_Symbol, _Period, 0));
   }
}

//+------------------------------------------------------------------+
void checkLineRemovalCondition()
{
   int period_3 = (int)MathCeil(i_SemaforPeriod_3),
       period_2 = (int)MathCeil(i_SemaforPeriod_2);
   static int scannedBar = MathMax(period_3, period_2);
   checkLineValidity();
}

//+------------------------------------------------------------------+
void scanSemafor(int p_lookBack)
{
   for(int i = p_lookBack; i > 0; i--)
   {
      sTSemaforDot currSemafor = getSemaforDot(PERIOD_CURRENT, i);
      if (currSemafor.period == 2 && !i_UsePeriod_2)
         continue;
      if (currSemafor.period == 3 && !i_UsePeriod_3)
         continue;
      if (currSemafor.period == 1)
         continue;
      if (  currSemafor.type == en_upperSemafor
            && isFoundOppositeSemafor(PERIOD_CURRENT, i, en_lowerSemafor, currSemafor.period)
            && isSemaforDotNotRecordedYet(currSemafor)
         )
      {
         createLine(currSemafor);
         addToEndOfArray(g_checkedSemafor, currSemafor);
      }
      /*----*/
      if (  currSemafor.type == en_lowerSemafor
            && isFoundOppositeSemafor(PERIOD_CURRENT, i, en_upperSemafor, currSemafor.period)
            && isSemaforDotNotRecordedYet(currSemafor)
         )
      {
         createLine(currSemafor);
         addToEndOfArray(g_checkedSemafor, currSemafor);
      }
   }
}

//+------------------------------------------------------------------+
void getCHOC_Direction(ENUM_TIMEFRAMES p_timeframe, int &p_direction, double &p_CHOC, datetime &p_CHOC_time,
                       double &p_upperBOS, double &p_lowerBOS)
{
   int lookFor = 0; /*-- flag to tell cycle to get first upper & lower point --*/
   sTSemaforDot upperCHOC, lowerCHOC;
   sTSemaforDot upperSemafore,
                lowerSemafore;
   /*----*/
   for(int i = iBars(_Symbol, p_timeframe) - 1; i > 0; i--)
   {
      sTSemaforDot currSemafor = getSemaforDot(p_timeframe, i);
      if (lookFor == 0 && (upperSemafore.price == 0 || lowerSemafore.price == 0))
      {
         if (currSemafor.type == en_lowerSemafor && currSemafor.period == 3 && lowerSemafore.price == 0)
            lowerSemafore = currSemafor;
         if (currSemafor.type == en_upperSemafor && currSemafor.period == 3 && upperSemafore.price == 0)
            upperSemafore = currSemafor;
         if (upperSemafore.price != 0 && lowerSemafore.price != 0)
            lookFor = -1;
         continue;
      }
      /*------------------------------------------------------------------*/
      /*-- updating after CHOC down --*/
      if (  lowerCHOC.price != 0
            && iClose(_Symbol, p_timeframe, i) < lowerCHOC.price )
      {
         lowerSemafore = lowerCHOC;
         int chocIndex = iBarShift(_Symbol, p_timeframe, lowerCHOC.time);
         int highestPointIndex = iHighest(_Symbol, p_timeframe, MODE_HIGH, chocIndex - i + 1, i);
         upperCHOC.price = iHigh(_Symbol, p_timeframe, highestPointIndex );
         p_CHOC = upperCHOC.price;
         upperCHOC.time = iTime(_Symbol, p_timeframe, highestPointIndex);
         p_CHOC_time = upperCHOC.time;
         lowerCHOC.price = 0;
         lowerCHOC.time = 0;
         p_direction = BEARISH;
      }
      /*-- updating after CHOC up --*/
      if (  upperCHOC.price != 0
            && iClose(_Symbol, p_timeframe, i) > upperCHOC.price )
      {
         upperSemafore = upperCHOC;
         int chocIndex = iBarShift(_Symbol, p_timeframe, upperCHOC.time);
         int lowestPointIndex = iLowest(_Symbol, p_timeframe, MODE_LOW, chocIndex - i + 1, i);
         lowerCHOC.price = iLow(_Symbol, p_timeframe, lowestPointIndex);
         p_CHOC = lowerCHOC.price;
         lowerCHOC.time = iTime(_Symbol, p_timeframe, lowestPointIndex);
         p_CHOC_time = lowerCHOC.time;
         upperCHOC.price = 0;
         upperCHOC.time = 0;
         p_direction = BULLISH;
      }
      /*------------------------------------------------------------------*/
      /*-- Upper BOS--*/
      if (  upperSemafore.price != 0
            && NormalizeDouble(iClose(_Symbol, p_timeframe, i), _Digits) > NormalizeDouble(upperSemafore.price, _Digits))
      {
         int prevSemaforHiIndex = iBarShift(_Symbol, p_timeframe, upperSemafore.time);
         int lowestPointIndex = iLowest(_Symbol, p_timeframe, MODE_LOW, prevSemaforHiIndex - i + 1, i);
         lowerCHOC.price = iLow(_Symbol, p_timeframe, lowestPointIndex);
         p_CHOC = lowerCHOC.price;
         lowerCHOC.time = iTime(_Symbol, p_timeframe, lowestPointIndex);
         p_CHOC_time = lowerCHOC.time;
         lowerSemafore.price = lowerCHOC.price;
         lowerSemafore.time = lowerCHOC.time;
         upperCHOC.price = 0;
         upperCHOC.time = 0;
         p_direction = BULLISH;
      }
      /*-- Lower BOS --*/
      if (  lowerSemafore.price != 0
            && NormalizeDouble(iClose(_Symbol, p_timeframe, i), _Digits) < NormalizeDouble(lowerSemafore.price, _Digits))
      {
         int prevSemaforLoIndex = iBarShift(_Symbol, p_timeframe, lowerSemafore.time);
         int highestPointIndex = iHighest(_Symbol, p_timeframe, MODE_HIGH, prevSemaforLoIndex - i + 1, i);
         upperCHOC.price = iHigh(_Symbol, p_timeframe, highestPointIndex );
         p_CHOC = upperCHOC.price;
         upperCHOC.time = iTime(_Symbol, p_timeframe, highestPointIndex);
         p_CHOC_time = upperCHOC.time;
         upperSemafore.price = upperCHOC.price;
         upperSemafore.time = upperCHOC.time;
         lowerCHOC.price = 0;
         lowerCHOC.time = 0;
         p_direction = BEARISH;
      }
      /*-- updating upper semafor --*/
      if (  p_direction == BULLISH
            && currSemafor.type == en_upperSemafor
            && currSemafor.period == 3
            && isFoundOppositeSemafor(p_timeframe, i, en_lowerSemafor, 3)
            && currSemafor.time > upperSemafore.time
            && currSemafor.price > upperSemafore.price
         )
      {
         upperSemafore = currSemafor;
         p_upperBOS = upperSemafore.price;
      }
      /*-- updating lower semafor --*/
      if (  p_direction == BEARISH
            && currSemafor.type == en_lowerSemafor
            && currSemafor.period == 3
            && isFoundOppositeSemafor(p_timeframe, i, en_upperSemafor, 3)
            && currSemafor.time > lowerSemafore.time
            && currSemafor.price < lowerSemafore.price
         )
      {
         lowerSemafore = currSemafor;
         p_lowerBOS = lowerSemafore.price;
      }
   }
}

//+------------------------------------------------------------------+
bool isSemaforDotNotRecordedYet(sTSemaforDot & p_semaforDot)
{
   if (ArraySize(g_checkedSemafor) <= 0)
      return true;
   /*----*/
   for(int i = ArraySize(g_checkedSemafor) - 1; i >= 0; i--)
   {
      if (  g_checkedSemafor[i].type == p_semaforDot.type
            && g_checkedSemafor[i].period == p_semaforDot.period
            && g_checkedSemafor[i].time == p_semaforDot.time
            && g_checkedSemafor[i].price == p_semaforDot.price
         )
         return false;
   }
   return true;
}

//+------------------------------------------------------------------+
void createLine(sTSemaforDot & p_semaforDot)
{
   string type = p_semaforDot.type == en_upperSemafor ? "upper_" : "lower_";
   string name = PREFIX + type + (string)p_semaforDot.time;
   color lineColor = p_semaforDot.type == en_upperSemafor ? clrOrangeRed : clrLimeGreen;
   createTrendLine(ChartID(), name, p_semaforDot.price, p_semaforDot.price, p_semaforDot.time, iTime(_Symbol, _Period, 0), lineColor, 0, 1, false);
}

//+------------------------------------------------------------------+
sTSemaforDot getSemaforDot(ENUM_TIMEFRAMES p_timeframe, int p_index)
{
   sTSemaforDot result;
   double buff_0 = getSemaforIndicatorValue(p_timeframe, 0, p_index),
          buff_1 = getSemaforIndicatorValue(p_timeframe, 1, p_index),
          buff_2 = getSemaforIndicatorValue(p_timeframe, 2, p_index),
          buff_3 = getSemaforIndicatorValue(p_timeframe, 3, p_index),
          buff_4 = getSemaforIndicatorValue(p_timeframe, 4, p_index),
          buff_5 = getSemaforIndicatorValue(p_timeframe, 5, p_index);
   if (buff_5 != 0 || buff_3 != 0 /*-- || buff_1 != 0 --*/)
   {
      result.type = en_upperSemafor;
      result.period = buff_5 != 0 ? 3 : buff_3 != 0 ? 2 : 1;
      result.price = iHigh(_Symbol, p_timeframe, p_index);
      result.time = iTime(_Symbol, p_timeframe, p_index);
   }
   /*----*/
   if (buff_4 != 0 || buff_2 != 0 /*-- || buff_0 != 0 --*/)
   {
      result.type = en_lowerSemafor;
      result.period = buff_4 != 0 ? 3 : buff_2 != 0 ? 2 : 1;
      result.price = iLow(_Symbol, p_timeframe, p_index);
      result.time = iTime(_Symbol, p_timeframe, p_index);
   }
   return result;
}

//+------------------------------------------------------------------+
bool isNoOpenAndPendingOrder()
{
   ST_OrderCount totalOrder;
   totalOrder.count(i_MagicNumber);
   if (totalOrder.buy + totalOrder.buystop + totalOrder.sell + totalOrder.sellstop <= 0)
      return true;
   return false;
}

//+------------------------------------------------------------------+
bool isNoOpenOrder()
{
   ST_OrderCount totalOrder;
   totalOrder.count(i_MagicNumber);
   if (totalOrder.buy + totalOrder.sell <= 0)
      return true;
   return false;
}

//+------------------------------------------------------------------+
void checkTradingCondition_ScaleInEntryModel()
{
   ST_SendOrderResult sendResult;
   ST_OrderCount orderTotal;
   orderTotal.count(i_MagicNumber);
   for(int i = ObjectsTotal(ChartID()) - 1; i >= 0; i--)
   {
      string name = ObjectName(ChartID(), i);
      if (ObjectGetInteger(ChartID(), name, OBJPROP_TYPE) != OBJ_TREND)
         continue;
      double price = ObjectGetDouble(ChartID(), name, OBJPROP_PRICE1);
      bool isAllowToTradeOnLevel = i_TradeCHOCLevel ||
                                   (!i_TradeCHOCLevel && NormalizeDouble(price, _Digits) != NormalizeDouble(g_chocLevel, _Digits));
      double distanceToChoc = MathAbs(price - g_chocLevel);
      bool isDistanceToCHOC_Allowed = (distanceToChoc < i_MaxDistanceToCHOC * _Point) || i_MaxDistanceToCHOC == 0;
      if (  isStringFind(name, "upper")
            && iClose(_Symbol, _Period, 1) > price
            && isAllowToOpenOrderAt(price)
            /*-- && (orderTotal.buy + orderTotal.sell <= 0) --*/
            && g_chochDirection == BEARISH
            && isAllowToTradeOnLevel
            && isDistanceToCHOC_Allowed
            && (orderTotal.buy + orderTotal.buylimit + orderTotal.buystop <= 0)
         )
      {
         closeAllPendingOrder();
         sendResult = openScaleInPendingOrder(SELL, price);
      }
      /*----*/
      if (  isStringFind(name, "lower")
            && iClose(_Symbol, _Period, 1) < price
            && isAllowToOpenOrderAt(price)
            /*-- && (orderTotal.buy + orderTotal.sell <= 0) --*/
            && g_chochDirection == BULLISH
            && isAllowToTradeOnLevel
            && isDistanceToCHOC_Allowed
            && (orderTotal.sell + orderTotal.sellstop + orderTotal.selllimit <= 0)
         )
      {
         closeAllPendingOrder();
         sendResult = openScaleInPendingOrder(BUY, price);
      }
   }
   if (sendResult.isSuccess && ArraySize(sendResult.ticket) > 0)
   {
      addToEndOfArray(g_managedPosition, sendResult.ticket[0]);
   }
}

//+------------------------------------------------------------------+
void checkTradingCondition_OriginalEntryModel()
{
   ST_SendOrderResult sendResult;
   ST_OrderCount orderTotal;
   orderTotal.count(i_MagicNumber);
   for(int i = ObjectsTotal(ChartID()) - 1; i >= 0; i--)
   {
      string name = ObjectName(ChartID(), i);
      if (ObjectGetInteger(ChartID(), name, OBJPROP_TYPE) != OBJ_TREND)
         continue;
      double price = ObjectGetDouble(ChartID(), name, OBJPROP_PRICE1);
      bool isAllowToTradeOnLevel = i_TradeCHOCLevel ||
                                   (!i_TradeCHOCLevel && NormalizeDouble(price, _Digits) != NormalizeDouble(g_chocLevel, _Digits));
      if (  isStringFind(name, "upper")
            && iClose(_Symbol, _Period, 1) > price
            && isAllowToOpenOrderAt(price)
            && (orderTotal.buy + orderTotal.sell <= 0)
            && g_chochDirection == BEARISH
            && isAllowToTradeOnLevel
         )
      {
         closeAllPendingOrder();
         sendResult = openPendingOrder(SELL, price);
         if (sendResult.isSuccess && ArraySize(sendResult.ticket) > 0)
            g_lastPosition = sendResult.ticket[0];
      }
      /*----*/
      if (  isStringFind(name, "lower")
            && iClose(_Symbol, _Period, 1) < price
            && isAllowToOpenOrderAt(price)
            && (orderTotal.buy + orderTotal.sell <= 0)
            && g_chochDirection == BULLISH
            && isAllowToTradeOnLevel
         )
      {
         closeAllPendingOrder();
         sendResult = openPendingOrder(BUY, price);
         if (sendResult.isSuccess && ArraySize(sendResult.ticket) > 0)
            g_lastPosition = sendResult.ticket[0];
      }
   }
   if (sendResult.isSuccess && ArraySize(sendResult.ticket) > 0)
   {
      addToEndOfArray(g_managedPosition, sendResult.ticket[0]);
   }
}

//+------------------------------------------------------------------+
void closeAllPendingOrder()
{
   closeOrderByType(OP_BUYSTOP, i_MagicNumber, i_MaxRetries, i_MaxSlippage);
   closeOrderByType(OP_BUYLIMIT, i_MagicNumber, i_MaxRetries, i_MaxSlippage);
   closeOrderByType(OP_SELLSTOP, i_MagicNumber, i_MaxRetries, i_MaxSlippage);
   closeOrderByType(OP_SELLLIMIT, i_MagicNumber, i_MaxRetries, i_MaxSlippage);
}

//+------------------------------------------------------------------+
bool isNoOpenBuyOrder()
{
   ST_OrderCount orderTotal;
   orderTotal.count(i_MagicNumber);
   if (orderTotal.buy <= 0)
      return true;
   else
      return false;
}

//+------------------------------------------------------------------+
bool isNoOpenSellOrder()
{
   ST_OrderCount orderTotal;
   orderTotal.count(i_MagicNumber);
   if (orderTotal.sell <= 0)
      return true;
   else
      return false;
}

//+------------------------------------------------------------------+
bool isAllowToOpenOrderAt(double p_price)
{
   for(int i = OrdersTotal() - 1; i >= 0; i--)
   {
      if (!OrderSelect(i, SELECT_BY_POS))
         continue;
      if (OrderSymbol() != _Symbol || OrderMagicNumber() != i_MagicNumber)
         continue;
      if (isStringFind(OrderComment(), DoubleToString(p_price, _Digits)))
         return false;
   }
   return true;
}

//+------------------------------------------------------------------+
// Check for new bar
bool isNewBar(const ENUM_TIMEFRAMES p_timeframe = PERIOD_CURRENT)
{
   datetime timeNow = iTime(_Symbol, p_timeframe, 0);
   static datetime newBar = timeNow;
   if (newBar == timeNow) return false;
   newBar = iTime(_Symbol, p_timeframe, 0);
   return true;
}


//+------------------------------------------------------------------+
bool closeOrderByTicket(   const int      p_ticket,
                           const int      p_slip,
                           const color    p_color,
                           int            p_trial,
                           string  p_caller = "")
{
   while(p_trial > 0 && IsExpertEnabled())
   {
      if(!OrderSelect(p_ticket, SELECT_BY_TICKET))
         return false;
      if (OrderType() <= 1)
      {
         if (OrderClose(p_ticket, OrderLots(), OrderClosePrice(), p_slip, p_color))
            return true;
         else
         {
            p_trial--;
            int err_code = GetLastError();
            if (errorSleep(err_code) > 0)
            {
               Sleep(errorSleep(err_code));
               p_caller = p_caller == "" ? __FUNCTION__ : p_caller;
               Print(p_caller, " | Error closing ticket ", p_ticket);
            }
            if (errorCritical(err_code)) return false;
         }
      }
      else if (OrderDelete(p_ticket, clrLimeGreen))
         return true;
      else
      {
         p_trial--;
         int err_code = GetLastError();
         if (errorSleep(err_code) > 0)
         {
            Sleep(errorSleep(err_code));
            p_caller = p_caller == "" ? __FUNCTION__ : p_caller;
            Print(p_caller, " | Error delete ticket ", p_ticket);
         }
         if (errorCritical(err_code)) return false;
      }
   }
   return false;
}

//+------------------------------------------------------------------+
bool isStringFind(string p_searchedTarget, string p_stringKeyWord)
{
   return StringFind(p_searchedTarget, p_stringKeyWord, 0) > -1;
}


//+------------------------------------------------------------------+
ST_OrderCount getOrderCount(const int &p_magicNumber)
{
   ST_OrderCount result;
   for(int i = 0; i < OrdersTotal(); i++)
   {
      if (!OrderSelect(i, SELECT_BY_POS)) continue;
      if (OrderSymbol() != _Symbol) continue;
      if (OrderMagicNumber() != p_magicNumber && p_magicNumber != 0) continue;
      if (OrderType() == OP_BUY)
         result.buy++;
      if (OrderType() == OP_SELL)
         result.sell++;
      if (OrderType() == OP_BUYSTOP)
         result.buystop++;
      if (OrderType() == OP_SELLSTOP)
         result.sellstop++;
      if (OrderType() == OP_BUYLIMIT)
         result.buylimit++;
      if (OrderType() == OP_SELLLIMIT)
         result.selllimit++;
   }
   return result;
}


//+------------------------------------------------------------------+
void closeAllOrder(  const int p_magicNumber,
                     const int p_slippage,
                     string p_caller = "")
{
   for(int i = OrdersTotal() - 1; i >= 0; i--)
   {
      if (!OrderSelect(i, SELECT_BY_POS)) continue;
      if ((OrderMagicNumber() != p_magicNumber && p_magicNumber != 0) || OrderSymbol() != _Symbol) continue;
      int retry = 3;
      bool result = false;
      while (result == false && retry > 0 && !IsStopped())
      {
         if (OrderType() > 1)
            result = OrderDelete(OrderTicket(), clrLime);
         else
            result = OrderClose(OrderTicket(), OrderLots(), OrderClosePrice(), p_slippage, clrLimeGreen);
         if (!result)
         {
            p_caller = p_caller == "" ? __FUNCTION__ : p_caller;
            Print(p_caller, " | Error closing order.", errorString(_LastError));
            Sleep(errorSleep(_LastError));
         }
         retry--;
      }
   }
}

//+------------------------------------------------------------------+
void closeOrderByType( const int p_type,
                       const int p_magicNumber,
                       const int p_retries,
                       const int p_slippage,
                       string p_caller = "")
{
   int total = OrdersTotal();
   for(int i = total - 1; i >= 0; i--)
   {
      if (!OrderSelect(i, SELECT_BY_POS)) continue;
      if ((OrderMagicNumber() != p_magicNumber && p_magicNumber != 0) || OrderSymbol() != _Symbol) continue;
      if (OrderType() != p_type) continue;
      int retry = p_retries;
      bool result = false;
      while (result == false && retry > 0 && !IsStopped())
      {
         if (OrderType() > 1)
            result = OrderDelete(OrderTicket(), clrLime);
         else
            result = OrderClose(OrderTicket(), OrderLots(), OrderClosePrice(), p_slippage, clrLimeGreen);
         if (!result)
         {
            p_caller = p_caller == "" ? __FUNCTION__ : p_caller;
            Print(p_caller, " | Error closing order.", errorString(_LastError));
            Sleep(errorSleep(_LastError));
         }
         retry--;
      }
   }
}

//+------------------------------------------------------------------+
template <typename array>
void addToEndOfArray(array &destination[], array &insertedElement)
{
   int size = ArraySize(destination);
   ArrayResize(destination, size + 1);
   destination[size] = insertedElement;
}

//+------------------------------------------------------------------+
template <typename array>
void deleteArrayElement(array &target[], int p_index)
{
   int size = ArraySize(target);
   for(int i = p_index; i < size - 1; i++)
   {
      target[i] = target[i + 1];
   }
   ArrayResize(target, size - 1);
}

//+------------------------------------------------------------------+
ST_SendOrderResult openScaleInPendingOrder(int p_OpType, double p_price)
{
   ST_Order order;
   ST_SendOrderResult sendOrderResult;
   order.expire = 0;
   if (p_OpType == BUY)
   {
      if (p_price < ASK)
         order.type = OP_BUYLIMIT;
      if (p_price > ASK)
         order.type = OP_BUYSTOP;
      if (p_price == ASK)
         order.type = OP_BUY;
   }
   if (p_OpType == SELL)
   {
      if (p_price < BID)
         order.type = OP_SELLSTOP;
      if (p_price > BID)
         order.type = OP_SELLLIMIT;
      if (p_price == BID)
         order.type = OP_SELL;
   }
   order.comment = DoubleToStr(p_price, _Digits);
   order.magic = i_MagicNumber;
   order.openprice = p_price;
   order.stoploss = getStoploss(order.type, order.openprice);
   order.size = getScaleInLotSize();
   order.openprice = NormalizeDouble(order.openprice, _Digits);
   order.stoploss = 0;
   order.takeprofit = 0;
   sendOrderResult = sendOrder(order, i_MaxSlippage, clrLimeGreen, i_MaxRetries);
   return sendOrderResult;
}

//+------------------------------------------------------------------+
double getScaleInLotSize()
{
   double result = 0;
   ST_OrderCount totalOrder;
   totalOrder.count(i_MagicNumber);
   /*-- not first order -> use initial lot size --*/
   if (totalOrder.buy + totalOrder.sell > 0)
   {
      result = i_LotSize;
   }
   /*-- first order -> calculate amount lot to recover previous losses ---*/
   else
   {
      double tickValue = MarketInfo(OrderSymbol(), MODE_TICKVALUE),
             point = MarketInfo(OrderSymbol(), MODE_POINT);
      double tpInDollar = i_LotSize * tickValue * i_TakeProfitBasket;
      double targetProfit = MathAbs(getTotalBasketDollarClosedPL()) + tpInDollar;
      /*--#del--*/
      PrintFormat("### total previous loss = %.2f", getTotalBasketDollarClosedPL());
      PrintFormat("### tp in dollar = %.2f", tpInDollar);
      PrintFormat("### target profit = %.2f", targetProfit);
      result = calculateOrderLot(targetProfit, 100, i_TakeProfitBasket, _Symbol);
   }
   return result;
}

//+------------------------------------------------------------------+
ST_SendOrderResult openPendingOrder(int p_OpType, double p_price)
{
   ST_Order order;
   ST_SendOrderResult sendOrderResult;
   order.expire = 0;
   if (p_OpType == BUY)
   {
      if (p_price < ASK)
         order.type = OP_BUYLIMIT;
      if (p_price > ASK)
         order.type = OP_BUYSTOP;
      if (p_price == ASK)
         order.type = OP_BUY;
   }
   if (p_OpType == SELL)
   {
      if (p_price < BID)
         order.type = OP_SELLSTOP;
      if (p_price > BID)
         order.type = OP_SELLLIMIT;
      if (p_price == BID)
         order.type = OP_SELL;
   }
   order.comment = DoubleToStr(p_price, _Digits);
   if (i_UsePlanB && isPlanB_Active())
      order.comment += order.comment + "#B";
   order.magic = i_MagicNumber;
   order.openprice = p_price;
   order.stoploss = getStoploss(order.type, order.openprice);
   double slLength = MathAbs(order.openprice - order.stoploss);
   double tpRR = getTakeProfitRR();
   if (i_TakeProfitRR != 0)
      order.takeprofit = (order.type == OP_BUYSTOP || order.type == OP_BUYLIMIT || order.type == OP_BUY) ? order.openprice + (tpRR * slLength)
                         : order.openprice - (tpRR * slLength);
   order.size = getLotSize(slLength);
   order.openprice = NormalizeDouble(order.openprice, _Digits);
   order.stoploss = NormalizeDouble(order.stoploss, _Digits);
   order.takeprofit = NormalizeDouble(order.takeprofit, _Digits);
   sendOrderResult = sendOrder(order, i_MaxSlippage, clrLimeGreen, i_MaxRetries);
   return sendOrderResult;
}

//+------------------------------------------------------------------+
ST_SendOrderResult openReEntryPendingOrder(int p_OpType, double p_price, double p_stoploss)
{
   ST_Order order;
   ST_SendOrderResult sendOrderResult;
   order.expire = 0;
   if (p_OpType == BUY)
   {
      if (p_price < ASK)
         order.type = OP_BUYLIMIT;
      if (p_price > ASK)
         order.type = OP_BUYSTOP;
      if (p_price == ASK)
         order.type = OP_BUY;
   }
   if (p_OpType == SELL)
   {
      if (p_price < BID)
         order.type = OP_SELLSTOP;
      if (p_price > BID)
         order.type = OP_SELLLIMIT;
      if (p_price == BID)
         order.type = OP_SELL;
   }
   order.comment = DoubleToStr(p_price, _Digits);
   order.magic = i_MagicNumber;
   order.openprice = p_price;
   order.stoploss = p_stoploss;
   double slLength = MathAbs(order.openprice - order.stoploss);
   double tpRR = getTakeProfitRR();
   order.takeprofit = (order.type == OP_BUYSTOP || order.type == OP_BUYLIMIT || order.type == OP_BUY) ? order.openprice + (tpRR * slLength)
                      : order.openprice - (tpRR * slLength);
   order.size = getLotSize(slLength);
   order.openprice = NormalizeDouble(order.openprice, _Digits);
   order.stoploss = NormalizeDouble(order.stoploss, _Digits);
   order.takeprofit = NormalizeDouble(order.takeprofit, _Digits);
   sendOrderResult = sendOrder(order, i_MaxSlippage, clrLimeGreen, i_MaxRetries);
   return sendOrderResult;
}

//+------------------------------------------------------------------+
double getTakeProfitRR()
{
   double result = 0;
   if (i_UseRecover)
   {
      result = getTotalPositionLossCount() == 0 ? i_TakeProfitRR :
               getTotalPositionLossCount() + 1;
      if (i_UsePlanA && getTotalPositionLossCount() >= i_PlanA_MaxLossTrade)
         result = i_PlanANewRR;
      if (i_UsePlanB && isPlanB_Active())
      {
         result = i_PlanANewRR + getTotalPlanBOpenedOrder() + 1;
      }
   }
   else
      result = i_TakeProfitRR;
   return result;
}

//+------------------------------------------------------------------+
double getTotalPlanBOpenedOrder()
{
   int total = 0;
   for(int i = ArraySize(g_managedPosition) - 1; i >= 0; i--)
   {
      if (OrderSelect(g_managedPosition[i], SELECT_BY_TICKET) == false)
         continue;
      if (OrderSymbol() != _Symbol || OrderMagicNumber() != i_MagicNumber || OrderType() > 1)
         continue;
      if (isStringFind(OrderComment(), "#B"))
         total++;
   }
   return total;
}

//+------------------------------------------------------------------+
bool isPlanB_Active()
{
   double totalClosedLot = getTotalClosedLot();
   return (totalClosedLot > i_PlanBLotSize);
}

//+------------------------------------------------------------------+
double getTotalClosedLot()
{
   double result = 0;
   for(int i = ArraySize(g_managedPosition) - 1; i >= 0; i--)
   {
      if (OrderSelect(g_managedPosition[i], SELECT_BY_TICKET) == false)
         continue;
      if (OrderType() > 1)
         continue;
      result += OrderLots();
   }
   return NormalizeDouble(result, 2);
}

//+------------------------------------------------------------------+
double getLotSize(double slLength)
{
   double result = 0;
   if (!i_UseRecover)
      result = calculateOrderLot(i_RiskPerTrade, 100, (slLength / _Point) - i_Commision, _Symbol);
   if (i_UseRecover)
   {
      if (!i_UsePlanA)
         result = calculateOrderLot(i_RiskPerTrade, 100, (slLength / _Point) - i_Commision, _Symbol);
      if (i_UsePlanA)
      {
         if (getTotalPositionLossCount() < i_PlanA_MaxLossTrade)
            result = calculateOrderLot(i_RiskPerTrade, 100, slLength / _Point - i_Commision, _Symbol);
         if (getTotalPositionLossCount() >= i_PlanA_MaxLossTrade)
         {
            double profitTarget = MathAbs(getTotalPositionDollarLoss()) + i_RiskPerTrade;
            result = calculateOrderLot(profitTarget, 100, (slLength / _Point * i_PlanANewRR) - i_Commision, _Symbol);
         }
      }
      /*----*/
      if (i_UsePlanB && isPlanB_Active())
      {
         double profitTarget = MathAbs(getTotalPositionDollarLoss()) + i_RiskPerTrade;
         result = calculateOrderLot(profitTarget, 100, (slLength / _Point * getTakeProfitRR()) - i_Commision, _Symbol);
      }
   }
   return result;
}

//+------------------------------------------------------------------+
double getTotalPositionDollarLoss()
{
   double total = 0;
   for(int i = ArraySize(g_managedPosition) - 1; i >= 0; i--)
   {
      if (OrderSelect(g_managedPosition[i], SELECT_BY_TICKET) == false)
         continue;
      if (OrderCloseTime() <= 0 || OrderProfit() > 0)
         continue;
      total += OrderProfit();
      total += OrderCommission();
      total += OrderSwap();
   }
   return total;
}

//+------------------------------------------------------------------+
double getStoploss(int p_type, double p_openPrice)
{
   double result = 0;
   if (p_type == OP_BUYSTOP || p_type == OP_BUYLIMIT || p_type == OP_BUY)
   {
      if (i_StoplossByZigzag)
         result = getLatestZigzag(MODE_LOWER);
      if (i_Stoploss != 0)
         result = p_openPrice - (i_Stoploss  * _Point);
   }
   /*----*/
   if (p_type == OP_SELLSTOP || p_type == OP_SELLLIMIT || p_type == OP_SELL)
   {
      if (i_StoplossByZigzag)
         result = getLatestZigzag(MODE_UPPER);
      if (i_Stoploss != 0)
         result = p_openPrice + (i_Stoploss  * _Point);
   }
   return result;
}

//+------------------------------------------------------------------+
double getLatestZigzag(int p_upperLower)
{
   double result = 0;
   int buffer = p_upperLower == MODE_UPPER ? 1 :
                p_upperLower == MODE_LOWER ? 2 : -1;
   for(int i = 0; i < iBars(_Symbol, _Period); i++)
   {
      result = iCustom(_Symbol, _Period, "Zigzag", i_ZigagDepth, i_ZigzagDeviation, i_ZigzagBacktstep, buffer, i);
      if (result != 0)
      {
         result = p_upperLower == MODE_UPPER ? iHigh(_Symbol, _Period, i) : iLow(_Symbol, _Period, i);
         break;
      }
   }
   return result;
}

//+------------------------------------------------------------------+
void checkNewSemafor()
{
   int period_3 = (int)MathCeil(i_SemaforPeriod_3),
       period_2 = (int)MathCeil(i_SemaforPeriod_2);
   static int scannedBar = MathMax(period_3, period_2) * 2;
   scanSemafor(scannedBar);
}

//+------------------------------------------------------------------+
void checkLineValidity()
{
   for(int j = ObjectsTotal(ChartID()); j >= 0; j--)
   {
      string objName = ObjectName(ChartID(), j);
      if (!isStringFind(objName, PREFIX))
         continue;
      if (ObjectGetInteger(ChartID(), objName, OBJPROP_TYPE) != OBJ_TREND)
         continue;
      double price = ObjectGetDouble(ChartID(), objName, OBJPROP_PRICE1);
      datetime time_1 = (datetime)ObjectGetInteger(ChartID(), objName, OBJPROP_TIME1);
      int shift = iBarShift(_Symbol, _Period, time_1);
      for(int i = shift - 1; i > 0; i--)
      {
         if (  isStringFind(objName, "upper")
               && iClose(_Symbol, _Period, i) > price
            )
         {
            deleteLineBelowPrice("upper", price, iTime(_Symbol, _Period, i));
         }
         /*----*/
         if (  isStringFind(objName, "lower")
               && iClose(_Symbol, _Period, i) < price
            )
         {
            deleteLineAbovePrice("lower", price, iTime(_Symbol, _Period, i));
         }
      }
   }
}

//+------------------------------------------------------------------+
void deleteLineBelowPrice(string p_identifier, double p_price, datetime p_candleTime)
{
   for(int i = ObjectsTotal() - 1; i >= 0; i--)
   {
      string objName = ObjectName(ChartID(), i);
      if (!isStringFind(objName, PREFIX))
         continue;
      if (!isStringFind(objName, p_identifier))
         continue;
      if (ObjectGetInteger(ChartID(), objName, OBJPROP_TYPE) != OBJ_TREND)
         continue;
      double price = ObjectGetDouble(ChartID(), objName, OBJPROP_PRICE1);
      datetime time_1 = (datetime)ObjectGetInteger(ChartID(), objName, OBJPROP_TIME1);
      if (  price < p_price
            && time_1 < p_candleTime
         )
         ObjectDelete(ChartID(), objName);
   }
}

//+------------------------------------------------------------------+
void deleteLineAbovePrice(string p_identifier, double p_price, datetime p_candleTime)
{
   for(int i = ObjectsTotal() - 1; i >= 0; i--)
   {
      string objName = ObjectName(ChartID(), i);
      if (!isStringFind(objName, PREFIX))
         continue;
      if (!isStringFind(objName, p_identifier))
         continue;
      if (ObjectGetInteger(ChartID(), objName, OBJPROP_TYPE) != OBJ_TREND)
         continue;
      double price = ObjectGetDouble(ChartID(), objName, OBJPROP_PRICE1);
      datetime time_1 = (datetime)ObjectGetInteger(ChartID(), objName, OBJPROP_TIME1);
      if (  price > p_price
            && time_1 < p_candleTime
         )
         ObjectDelete(ChartID(), objName);
   }
}

//+------------------------------------------------------------------+
bool isFoundOppositeSemafor(ENUM_TIMEFRAMES p_timeframe, int p_index, eNSemaforType p_searchedSemafor, int p_semaforPeriod)
{
   for(int i = p_index - 1; i > 0; i--)
   {
      sTSemaforDot currSemafor = getSemaforDot(p_timeframe, i);
      if (currSemafor.type == p_searchedSemafor)
      {
         if (p_semaforPeriod == 2 && (currSemafor.period == 2 || currSemafor.period == 3))
            return true;
         if (p_semaforPeriod == 3 && currSemafor.period == 3)
            return true;
      }
   }
   return false;
}

//+------------------------------------------------------------------+
double getSemaforIndicatorValue(ENUM_TIMEFRAMES p_timeframe, int p_buffer, int p_index, string p_caller = "")
{
   ResetLastError();
   string indiName = INDICATOR_NAME;
   double value = iCustom(_Symbol, p_timeframe, indiName,
                          i_SemaforPeriod_1,
                          i_SemaforPeriod_2,
                          i_SemaforPeriod_3,
                          i_SemaforDevStep_1,
                          i_SemaforDevStep_2,
                          i_SemaforDevStep_3,
                          p_buffer,
                          p_index);
   int error = GetLastError();
   if (error > 1)
   {
      p_caller = p_caller == "" ? __FUNCTION__ : p_caller;
      Alert("Indicator " + indiName + " error. Error message : ", errorString(error));
      Print(p_caller, " | Error with error code : ", error);
      Print(p_caller, " | Error message : ", errorString(error));
      ExpertRemove();
   }
   return value;
}

//+------------------------------------------------------------------+
bool createTrendLine  (     const long     p_chartID,                          // chart's ID
                            const string   p_name,                             // line name
                            const double   p_price1,                           // line price 1
                            const double   p_price2,                           // line price 2
                            const datetime p_date1,                            // date 1
                            const datetime p_date2,                            // date 2
                            const color    p_color,                            // line color
                            const ENUM_LINE_STYLE p_style = STYLE_SOLID,       // line style
                            const int      p_width = 1,                        // line width
                            const bool     p_selectable = true,                // line selectable
                            const bool     p_backPosition = false,             // in the background
                            const bool     p_ray = false                       // ray
                      )
{
   ResetLastError();
   if (ObjectFind(p_name) == -1)
      ObjectCreate(p_chartID, p_name, OBJ_TREND, 0, 0, 0);
   ObjectSetDouble(p_chartID, p_name, OBJPROP_PRICE1, p_price1);
   ObjectSetDouble(p_chartID, p_name, OBJPROP_PRICE2, p_price2);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_TIME1, p_date1);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_TIME2, p_date2);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_STYLE, p_style);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_WIDTH, p_width);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_COLOR, p_color);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_BACK, p_backPosition);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_SELECTABLE, p_selectable);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_RAY, p_ray);
   ObjectSetInteger(p_chartID, p_name, OBJPROP_HIDDEN, true);
   ObjectSetString(p_chartID, p_name, OBJPROP_TOOLTIP, "\n");
   return true;
}

//+------------------------------------------------------------------+
ST_SendOrderResult sendOrder
(
   const ST_Order    &p_order,
   const int         p_slip,
   const color       p_clr,
   const int         p_retry,
   const bool        p_ecnMode = false,
   const bool        p_sound = false,
   string p_caller = ""
)
{
   ST_SendOrderResult orderResult;
   p_caller = p_caller == "" ? __FUNCTION__ : p_caller;
   if (  !isNotNegative(p_order.openprice, "Open price is not valid")
         || !isNotNegative(p_order.expire, "Expiration is not valid")
         || !isNotNegative(p_order.stoploss, "Stoploss is not valid")
         || !isNotNegative(p_order.takeprofit, "Takeprofit is not valid")
         || !isNotNegative(p_order.magic, "Magic number is not valid")
         || !isNotNegative(p_retry, "Retry variable is not valid")
         || !isNotNegative(p_slip, "Slippage is not valid")
         || !isValid(p_order.type, "Ordertype is not valid")
      )
   {
      orderResult.isSuccess = false;
      return orderResult;
   }
//---
   orderResult.errCode = 0;
   double maxLot = MarketInfo(p_order.symbol, MODE_MAXLOT);
   double minLot = MarketInfo(p_order.symbol, MODE_MINLOT);
   double opPrice = 0;
   int result = -1;
   int i;
   if (p_order.size <= maxLot)
   {
      for (i = 0; i < p_retry; i++)
      {
         RefreshRates();
         if (p_order.type == OP_BUY)
         {
            opPrice = p_order.openprice == 0 ? MarketInfo(p_order.symbol, MODE_ASK) :
                      p_order.openprice;
         }
         if (p_order.type == OP_SELL)
         {
            opPrice = p_order.openprice == 0 ? MarketInfo(p_order.symbol, MODE_BID) :
                      p_order.openprice;
         }
         if (p_order.openprice != 0 && p_order.type > 1) opPrice = p_order.openprice;
         if (!p_ecnMode) result = OrderSend(
                                        p_order.symbol,
                                        p_order.type, p_order.size,
                                        opPrice,
                                        p_slip,
                                        p_order.stoploss,
                                        p_order.takeprofit,
                                        p_order.comment,
                                        p_order.magic,
                                        p_order.expire,
                                        p_clr
                                     );
         if (p_ecnMode) result = OrderSend(
                                       p_order.symbol,
                                       p_order.type, p_order.size,
                                       opPrice,
                                       p_slip,
                                       0,
                                       0,
                                       p_order.comment,
                                       p_order.magic,
                                       0,
                                       p_clr
                                    );
         if (result != -1)
         {
            orderResult.lastOPtime = TimeCurrent();
            int size = ArraySize(orderResult.ticket);
            ArrayResize(orderResult.ticket, size + 1);
            orderResult.ticket[size] = result;
            if (p_ecnMode)
            {
               if (OrderSelect(result, SELECT_BY_TICKET, MODE_TRADES))
               {
                  if ((p_order.stoploss != 0 || p_order.takeprofit != 0) && !OrderModify(result, OrderOpenPrice(), p_order.stoploss, p_order.takeprofit, p_order.expire, p_clr))
                  {
                     orderResult.errCode = GetLastError();
                     if (orderResult.errCode > 1)
                     {
                        Print(__FUNCTION__, " | Failed to modify order. Error =" + errorString(orderResult.errCode));
                        if (p_sound) PlaySound(errorSound(orderResult.errCode));
                     }
                  }
               }
               else
               {
                  orderResult.errCode = GetLastError();
                  if (orderResult.errCode > 1)
                  {
                     Print(__FUNCTION__, " | Failed to select order. Error =" + errorString(orderResult.errCode));
                     if (p_sound) PlaySound(errorSound(orderResult.errCode));
                  }
               }
            }
            if (p_sound )PlaySound("ok.wav");
            Sleep(200);
            break;
         }
         else
         {
            orderResult.errCode = GetLastError();
            if (orderResult.errCode > 1)
            {
               Print(__FUNCTION__, " | Failed to open order. Error =" + errorString(orderResult.errCode));
               if (p_sound) PlaySound(errorSound(orderResult.errCode));
               if (errorSleep(orderResult.errCode) > 0) Sleep(errorSleep(orderResult.errCode));
               if (errorCritical(orderResult.errCode))
               {
                  orderResult.isSuccess = false;
                  return orderResult;
               }
            }
         }
      }
   }
   else
   {
      double remainingLots = p_order.size;
      double sendLots = 0;
      while(remainingLots > 0)
      {
         if (remainingLots > maxLot) sendLots = maxLot;
         if (remainingLots < maxLot) sendLots = remainingLots;
         for (i = 0; i <= p_retry; i++)
         {
            RefreshRates();
            if (p_order.openprice == 0 && p_order.type == OP_BUY) opPrice = MarketInfo(p_order.symbol, MODE_ASK);
            if (p_order.openprice == 0 && p_order.type == OP_SELL) opPrice = MarketInfo(p_order.symbol, MODE_BID);
            if (p_order.openprice != 0 && p_order.type > 1) opPrice = p_order.openprice;
            if (!p_ecnMode) result = OrderSend(
                                           p_order.symbol,
                                           p_order.type,
                                           sendLots,
                                           opPrice,
                                           p_slip,
                                           p_order.stoploss,
                                           p_order.takeprofit,
                                           p_order.comment,
                                           p_order.magic,
                                           p_order.expire,
                                           p_clr
                                        );
            if (p_ecnMode) result = OrderSend(
                                          p_order.symbol,
                                          p_order.type,
                                          sendLots,
                                          opPrice,
                                          p_slip,
                                          0,
                                          0,
                                          p_order.comment,
                                          p_order.magic,
                                          0,
                                          p_clr
                                       );
            if (result != -1)
            {
               orderResult.lastOPtime = TimeCurrent();
               int size = ArraySize(orderResult.ticket);
               ArrayResize(orderResult.ticket, size + 1);
               orderResult.ticket[size] = result;
               if (p_ecnMode)
               {
                  if (OrderSelect(result, SELECT_BY_TICKET, MODE_TRADES))
                  {
                     if (!OrderModify(result, OrderOpenPrice(), p_order.stoploss, p_order.takeprofit, p_order.expire, p_clr))
                     {
                        orderResult.errCode = GetLastError();
                        if (orderResult.errCode > 1)
                        {
                           Print(__FUNCTION__, " | Failed to modify order. Error =" + errorString(orderResult.errCode));
                           PlaySound(errorSound(orderResult.errCode));
                        }
                     }
                  }
                  else
                  {
                     orderResult.errCode = GetLastError();
                     if (orderResult.errCode > 1)
                     {
                        Print(__FUNCTION__, " | Failed to select order. Error =" + errorString(orderResult.errCode));
                        if (p_sound) PlaySound(errorSound(orderResult.errCode));
                     }
                  }
               }
               remainingLots = remainingLots - sendLots;
               Sleep(200);
               if (p_sound) PlaySound("ok.wav");
               break;
            }
            else
            {
               orderResult.errCode = GetLastError();
               {
                  Print(__FUNCTION__, " | Failed to open order. Error =" + errorString(orderResult.errCode));
                  if (p_sound) PlaySound(errorSound(orderResult.errCode));
                  if (errorSleep(orderResult.errCode) > 0) Sleep(errorSleep(orderResult.errCode));
                  if (errorCritical(orderResult.errCode))
                  {
                     orderResult.isSuccess = false;
                     return orderResult;
                  }
               }
            }
         }
      }
   }
   orderResult.isSuccess = true;
   return orderResult;
}

//+------------------------------------------------------------------+
template <typename varType>
bool isNotNegative(const varType pv_data, const string pv_message)
{
   if(pv_data < 0)
   {
      Print(pv_message + " | Error, ", typename (pv_data), " variable is negative.");
      return false;
   }
   else return true;
}

//+------------------------------------------------------------------+
bool isValid(const ENUM_TIMEFRAMES pv_tf, const  string pv_message)
{
   switch(pv_tf)
   {
   case  PERIOD_M1:
      return true;
      break;
   case  PERIOD_M5:
      return true;
      break;
   case  PERIOD_M15:
      return true;
      break;
   case  PERIOD_M30:
      return true;
      break;
   case  PERIOD_H1:
      return true;
      break;
   case  PERIOD_H4:
      return true;
      break;
   case  PERIOD_D1:
      return true;
      break;
   case  PERIOD_W1:
      return true;
      break;
   case  PERIOD_MN1:
      return true;
      break;
   case  PERIOD_CURRENT:
      return true;
      break;
   }
   Print(pv_message + " | Error, timeframe variable is invalid.");
   return false;
}

//+------------------------------------------------------------------+
bool isValid(const int v_orderType, const  string v_message)
{
   switch(v_orderType)
   {
   case OP_BUY :
      return true;
      break;
   case OP_SELL :
      return true;
      break;
   case OP_BUYSTOP :
      return true;
      break;
   case OP_SELLSTOP :
      return true;
      break;
   case OP_BUYLIMIT :
      return true;
      break;
   case OP_SELLLIMIT :
      return true;
      break;
   }
   Print(v_message + " | Error, order type is invalid.");
   return false;
}

//+------------------------------------------------------------------+
string errorString(int errorCode)
{
   switch(errorCode)
   {
   case 0:
      return("No error returned.");
   case 1:
      return("No error returned, but the result is unknown.");
   case 2:
      return("Common error.");
   case 3:
      return("Invalid trade parameters.");
   case 4:
      return("Trade server is busy.");
   case 5:
      return("Old version of the client terminal.");
   case 6:
      return("No connection with trade server.");
   case 7:
      return("Not enough rights.");
   case 8:
      return("Too frequent requests.");
   case 9:
      return("Malfunction trade operation.");
   case 64:
      return("Account disabled.");
   case 65:
      return("Invalid account.");
   case 128:
      return("Trade timeout.");
   case 129:
      return("Invalide price.");
   case 130:
      return("Invalid stops.");
   case 131:
      return("Invalid trade volume.");
   case 132:
      return("Market is closed.");
   case 133:
      return("Trade is disabled.");
   case 134:
      return("Not enough money.");
   case 135:
      return("Price changed.");
   case 136:
      return("Off quotes.");
   case 137:
      return("Broker is busy.");
   case 138:
      return("Requote.");
   case 139:
      return("Order is locked.");
   case 140:
      return("Long positions only allowed");
   case 141:
      return("Too many requests");
   case 145:
      return("Modification denied because order too close to market.");
   case 146:
      return("Trade context is busy.");
   case 147:
      return("Expirations are denied by broker");
   case 148:
      return("The amount of open and pending orders has reached the limit set by the broker");
   case 149:
      return("An attempt to open an order opposite to the existing one when hedging is disabled");
   case 150:
      return("An attempt to close an order contravening the FIFO rule");
//--- Runtime Error
   case 4000:
      return("No error returned");
   case 4001:
      return("Wrong function pointer");
   case 4002:
      return("Array index is out of range");
   case 4003:
      return("No memory for function call stack");
   case 4004:
      return("Recursive stack overflow");
   case 4005:
      return("Not enough stack for parameter");
   case 4006:
      return("No memory for parameter string");
   case 4007:
      return("No memory for temp string");
   case 4008:
      return("Not initialized string");
   case 4009:
      return("Not initialized string in array");
   case 4010:
      return("No memory for array string");
   case 4011:
      return("Too long string");
   case 4012:
      return("Remainder from zero divide");
   case 4013:
      return("Zero divide");
   case 4014:
      return("Unknown command");
   case 4015:
      return("Wrong jump (never generated error)");
   case 4016:
      return("Not initialized array");
   case 4017:
      return("DLL calls are not allowed");
   case 4018:
      return("Cannot load library");
   case 4019:
      return("Cannot call function");
   case 4020:
      return("Expert function calls are not allowed");
   case 4021:
      return("Not enough memory for temp string returned from function");
   case 4022:
      return("System is busy (never generated error)");
   case 4023:
      return("DLL-function call critical error");
   case 4024:
      return("Internal error");
   case 4025:
      return("Out of memory");
   case 4026:
      return("Invalid pointer");
   case 4027:
      return("Too many formatters in the format function");
   case 4028:
      return("Parameters count exceeds formatters count");
   case 4029:
      return("Invalid array");
   case 4030:
      return("No reply from chart");
   case 4050:
      return("Invalid function parameters count");
   case 4051:
      return("Invalid function parameter value");
   case 4052:
      return("String function internal error");
   case 4053:
      return("Some array error");
   case 4054:
      return("Incorrect series array using");
   case 4055:
      return("Custom indicator error");
   case 4056:
      return("Arrays are incompatible");
   case 4057:
      return("Global variables processing error");
   case 4058:
      return("Global variable not found");
   case 4059:
      return("Function is not allowed in testing mode");
   case 4060:
      return("Function is not allowed for call");
   case 4061:
      return("Send mail error");
   case 4062:
      return("String parameter expected");
   case 4063:
      return("Integer parameter expected");
   case 4064:
      return("Double parameter expected");
   case 4065:
      return("Array as parameter expected");
   case 4066:
      return("Requested history data is in updating state");
   case 4067:
      return("Internal trade error");
   case 4068:
      return("Resource not found");
   case 4069:
      return("Resource not supported");
   case 4070:
      return("Duplicate resource");
   case 4071:
      return("Custom indicator cannot initialize");
   case 4072:
      return("Cannot load custom indicator");
   case 4073:
      return("No history data");
   case 4074:
      return("No memory for history data");
   case 4075:
      return("Not enough memory for indicator calculation");
   case 4099:
      return("End of file");
   case 4100:
      return("Some file error");
   case 4101:
      return("Wrong file name");
   case 4102:
      return("Too many opened files");
   case 4103:
      return("Cannot open file");
   case 4104:
      return("Incompatible access to a file");
   case 4105:
      return("No order selected");
   case 4106:
      return("Unknown symbol");
   case 4107:
      return("Invalid price");
   case 4108:
      return("Invalid ticket");
   case 4109:
      return("Trade is not allowed. Enable checkbox 'Allow live trading' in the Expert Advisor properties");
   case 4110:
      return("Longs are not allowed. Check the Expert Advisor properties");
   case 4111:
      return("Shorts are not allowed. Check the Expert Advisor properties");
   case 4112:
      return("Automated trading by Expert Advisors/Scripts disabled by trade server");
   case 4200:
      return("Object already exists");
   case 4201:
      return("Unknown object property");
   case 4202:
      return("Object does not exist");
   case 4203:
      return("Unknown object type");
   case 4204:
      return("No object name");
   case 4205:
      return("Object coordinates error");
   case 4206:
      return("No specified subwindow");
   case 4207:
      return("Graphical object error");
   case 4210:
      return("Unknown chart property");
   case 4211:
      return("Chart not found");
   case 4212:
      return("Chart subwindow not found");
   case 4213:
      return("Chart indicator not found");
   case 4220:
      return("Symbol select error");
   case 4250:
      return("Notification error");
   case 4251:
      return("Notification parameter error");
   case 4252:
      return("Notifications disabled");
   case 4253:
      return("Notification send too frequent");
   case 4260:
      return("FTP server is not specified");
   case 4261:
      return("FTP login is not specified");
   case 4262:
      return("FTP connection failed");
   case 4263:
      return("FTP connection closed");
   case 4264:
      return("FTP path not found on server");
   case 4265:
      return("File not found in the MQL4'\'Files directory to send on FTP server");
   case 4266:
      return("Common error during FTP data transmission");
   case 5001:
      return("Too many opened files");
   case 5002:
      return("Wrong file name");
   case 5003:
      return("Too long file name");
   case 5004:
      return("Cannot open file");
   case 5005:
      return("Text file buffer allocation error");
   case 5006:
      return("Cannot delete file");
   case 5007:
      return("Invalid file handle (file closed or was not opened)");
   case 5008:
      return("Wrong file handle (handle index is out of handle table)");
   case 5009:
      return("File must be opened with FILE_WRITE flag");
   case 5010:
      return("File must be opened with FILE_READ flag");
   case 5011:
      return("File must be opened with FILE_BIN flag");
   case 5012:
      return("File must be opened with FILE_TXT flag");
   case 5013:
      return("File must be opened with FILE_TXT or FILE_CSV flag");
   case 5014:
      return("File must be opened with FILE_CSV flag");
   case 5015:
      return("File read error");
   case 5016:
      return("File write error");
   case 5017:
      return("String size must be specified for binary file");
   case 5018:
      return("Incompatible file (for string arrays-TXT, for others-BIN)");
   case 5019:
      return("File is directory not file");
   case 5020:
      return("File does not exist");
   case 5021:
      return("File cannot be rewritten");
   case 5022:
      return("Wrong directory name");
   case 5023:
      return("Directory does not exist");
   case 5024:
      return("Specified file is not directory");
   case 5025:
      return("Cannot delete directory");
   case 5026:
      return("Cannot clean directory");
   case 5027:
      return("Array resize error");
   case 5028:
      return("String resize error");
   case 5029:
      return("Structure contains strings or dynamic arrays");
   case 5200:
      return("Invalid URL");
   case 5201:
      return("Failed to connect to specified URL");
   case 5202:
      return("Timeout exceeded");
   case 5203:
      return("HTTP request failed");
   }
   return("Unknown Error Code.");
}

//+------------------------------------------------------------------+
string errorSound(int errorCode)
{
   switch(errorCode)
   {
   case 0:
      return("timeout.wav");
   case 1:
      return("timeout.wav");
   case 2:
      return("timeout.wav");
   case 3:
      return("timeout.wav");
   case 4:
      return("timeout.wav");
   case 5:
      return("timeout.wav");
   case 6:
      return("timeout.wav");
   case 7:
      return("timeout.wav");
   case 8:
      return("timeout.wav");
   case 9:
      return("timeout.wav");
   case 64:
      return("timeout.wav");
   case 65:
      return("timeout.wav");
   case 128:
      return("timeout.wav");
   case 129:
      return("timeout.wav");
   case 130:
      return("timeout.wav");
   case 131:
      return("timeout.wav");
   case 132:
      return("timeout.wav");
   case 133:
      return("timeout.wav");
   case 134:
      return("timeout.wav");
   case 135:
      return("timeout.wav");
   case 136:
      return("timeout.wav");
   case 137:
      return("timeout.wav");
   case 138:
      return("timeout.wav");
   case 139:
      return("timeout.wav");
   case 140:
      return("timeout.wav");
   case 141:
      return("timeout.wav");
   case 145:
      return("timeout.wav");
   case 146:
      return("timeout.wav");
   }
   return("timeout.wav");
}

//+------------------------------------------------------------------+
int errorSleep(int errorCode)
{
   switch(errorCode)
   {
   case 0:
      return(0);//return("No error returned.");
   case 1:
      return(0);//return("No error returned, but the result is unknown.");
   case 2:
      return(0);//return("Common error.");
   case 3:
      return(0);//return("Invalid trade parameters.");
   case 4:
      return(2000);//return("Trade server is busy.");
   case 5:
      return(0);//return("Old version of the client terminal.");
   case 6:
      return(0);//return("No connection with trade server.");
   case 7:
      return(0);//return("Not enough rights.");
   case 8:
      return(3000);//return("Too frequent requests.");
   case 9:
      return(0);//return("Malfunction trade operation.");
   case 64:
      return(0);//return("Account disabled.");
   case 65:
      return(0);//return("Invalid account.");
   case 128:
      return(0);//return("Trade timeout.");
   case 129:
      return(0);//return("Invalide price.");
   case 130:
      return(0);//return("Invalid stops.");
   case 131:
      return(0);//return("Invalid trade volume.");
   case 132:
      return(0);//return("Market is closed.");
   case 133:
      return(0);//return("Trade is disabled.");
   case 134:
      return(0);//return("Not enough money.");
   case 135:
      return(250);//return("Price changed.");
   case 136:
      return(250);//return("Off quotes.");
   case 137:
      return(0);//return("Broker is busy.");
   case 138:
      return(250);//{ return("Requote."); PlaySound("Requote.wav");}
   case 139:
      return(0);//return("Order is locked.");
   case 140:
      return(0);//return("Long positions only allowed");
   case 141:
      return(3000);//return("Too many requests");
   case 145:
      return(3000);//return("Modification denied because order too close to market.");
   case 146:
      return(2000);//return("Trade context is busy.");
   }
   return(0);
}

//+------------------------------------------------------------------+
bool errorCritical(int errorCode)
{
   switch(errorCode)
   {
//----Non Critical Error--------------------------------------------------------------
   case 0:  //return("No error returned.");
   case 1:  //return("No error returned, but the result is unknown.");
   case 2:  //return("Common error.");
   case 4:  //return("Trade server is busy.");
   case 8:  //return("Too frequent requests.");
   case 128:   //return("Trade timeout.");
   case 129:   //return("Invalide price.");
   case 135:   //return("Price changed.");
   case 136:   //return("Off quotes.");
   case 137:   //return("Broker is busy.");
   case 138:   //{ return("Requote."); PlaySound("Requote.wav");}
   case 141:   //return("Too many requests");
   case 145:   //return("Modification denied because order too close to market.");
   case 146:
      return false;//return("Trade context is busy.");
//----Critical Error-------------------------------------------------------------------
   case 3:  //return("Invalid trade parameters.");
   case 6:  //return("No connection with trade server.");
   case 7:  //return("Not enough rights.");
   case 9:  //return("Malfunction trade operation.");
   case 64:  //return("Account disabled.");
   case 65:  //return("Invalid account.");
   case 130:  //return("Invalid stops.");
   case 131:  //return("Invalid trade volume.");
   case 132:  //return("Market is closed.");
   case 133:  //return("Trade is disabled.");
   case 134:  //return("Not enough money.");
   case 139:  //return("Order is locked.");
   case 140:
      return true;//return("Long positions only allowed");
   }
   return true;
}
//+------------------------------------------------------------------+
